/* audispd.c -- 
 * Copyright 2005,2006 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 *   Junji Kanemaru <junji.kanemaru@linuon.com>
 *
 */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <locale.h>

#include "libaudit.h"
#include "private.h"
#include "application.h"
#include "componentconfigurator.h"

// Local data
static int fault_occurred = 0;
static volatile int signaled = 0;
static int pipe_fd;

// Local functions
static int event_loop(void);

// SIGTERM handler
static void term_handler( int sig )
{
	signaled = 1;
}

int has_fault_occurred(void)
{
	return fault_occurred;
}

/*
 * main
 */
int main(int argc, char *argv[])
{
	struct sigaction sa;

	setlocale (LC_ALL, "");
	set_aumessage_mode(MSG_SYSLOG, DBG_NO);
	openlog("audispd", LOG_PID, LOG_DAEMON);
	audit_msg(LOG_NOTICE, "audisp starting...");

#ifndef DEBUG
	// Make sure we are root
	if (getuid() != 0) {
		audit_msg(LOG_ERR, "You must be root to run this program.");
		return 4;
	}
#endif

	// register sighandlers 
	sa.sa_flags = 0 ;
	sa.sa_handler = term_handler;
	sigemptyset( &sa.sa_mask ) ;
	sigaction( SIGTERM, &sa, NULL );
	sa.sa_handler = term_handler;
	sigemptyset( &sa.sa_mask ) ;
	sigaction( SIGCHLD, &sa, NULL );
	sa.sa_handler = SIG_IGN;
	sigaction( SIGHUP, &sa, NULL );
	(void)chdir("/");

	// change over to pipe_fd
	pipe_fd = dup(0);
	close(0);
	open("/dev/null", O_RDONLY);
	fcntl(pipe_fd, F_SETFD, FD_CLOEXEC);

	// Init plugins
	if (ComponentConfigurator::instance()->init()) {
		return 1;
	}

	// Start the program
	Application::instance()->resume();
	return event_loop();
}

// NOTE: This will be deleted since its handled by a builtin plugin
static int event_loop(void)
{
	void *data;
	struct iovec vec[2];
	struct audit_dispatcher_header hdr;

	// allocate data structures
	data = malloc(MAX_AUDIT_MESSAGE_LENGTH);
	if (data == NULL) {
		audit_msg(LOG_ERR, "Cannot allocate buffer");
		return 1;
	}
	memset(data, 0, MAX_AUDIT_MESSAGE_LENGTH);
	memset(&hdr, 0, sizeof(hdr));

	do {
		int rc;
		struct timeval tv;
		fd_set fd;

		FD_ZERO(&fd);
		FD_SET(pipe_fd, &fd);
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		rc = select(pipe_fd+1, &fd, NULL, NULL, &tv);
		if (rc == 0) 
			continue;
		 else if (rc < 0)
			break;

		// Get header first. it is fixed size 
		vec[0].iov_base = (void*)&hdr;
		vec[0].iov_len = sizeof(hdr);

        /* Load payload. always try max size. don't worry, it won't block */
        /* TODO: If we are going to do real streaming then this wouldn't work.
        ** It is because, since we so non-blocking I/O, the payload data 
        ** could be fragmented in some case, i.e. slow line etc.
        ** So we will have to read the pipe until we get exact size that
        ** header says, then there's no advantage to use readv().. */
		vec[1].iov_base = data;
		vec[1].iov_len = MAX_AUDIT_MESSAGE_LENGTH; 

		rc = readv(pipe_fd, vec, 2);
		if (rc <= 0) {
			audit_msg(LOG_ERR, "rc == %d(%s)", rc, strerror(errno));
			break;
		}

		// XXX dispatch events here XXX
		audit_msg(LOG_NOTICE,"audispd: type=%d, payload size=%d", 
			hdr.type, hdr.size);
		audit_msg(LOG_NOTICE,"audispd: data=\"%.*s\"", hdr.size,
			(char *)data);

	} while(!signaled);

	return 0;
}

