/*
 * circularqueue.cpp
 * Copyright 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Steve Grubb <sgrubb@redhat.com>
 *     Junji Kanemaru <junji.kanemaru@linuon.com>
 */
#include "circularqueue.h"
#include "event.h"
#include "fault.h"
#include <assert.h>
#include <cassert>

using namespace std;

/*
** Base CircularQueue class
*/

/**
 * Constructors/Destructors
 */
CircularQueue::CircularQueue(size_t maxSize) {
    this->maxSize = maxSize;
    size = 0;
    head = tail = 0;
    array = new Event[maxSize];
    pthread_mutex_init(&mutex, NULL);
}

CircularQueue::~CircularQueue() {
    Event evt;
    delete[] array;
    pthread_mutex_destroy(&mutex);
}

/**
 * This method places event data into the queue. The array of events are 
 * already created. This means that an assignment operator will be used to
 * add the event to the array. 
 */
void CircularQueue::enqueue (const Event& evt) {
    pthread_mutex_lock(&mutex);

    if (size == maxSize && head == tail) {
        head++;
        if (head == maxSize) {
            head = 0;
        }
    }

    /* Set Event */
    array[tail++] = evt;

    if (tail == maxSize) {
        tail = 0;
    }

    size++;
    if (size > maxSize) {
        size = maxSize;
    }

    pthread_mutex_unlock(&mutex);
}

/**
 * This method removes data from the queue. Because of concerns regarding 
 * dynamic creation and deletion of classes in a long running server, 
 * this will effectivly just zero out the event after copying it to a stack 
 * variable.
 */
bool CircularQueue::dequeue (Event& evt) {
    if (size == 0 && tail == head) {
        return false;
    }

    pthread_mutex_lock(&mutex);

    evt = array[head];
    if (evt == Value::VOID_VALUE) {
        pthread_mutex_unlock(&mutex);
        return false;
    }

    /* Clear element with Value::VOID_VALUE to make it de-referenced. 
    ** We need to do this otherwise instance will remain in the queue and, 
    ** hence, the Value data instatnce sits in application heap until 
    ** it gets overwrtten next time. */
    array[head++] = Value::VOID_VALUE;

    if (head == maxSize) {
        head = 0;
    }

    size--;

    pthread_mutex_unlock(&mutex);

    return true;
}


/*
** CircularPriorityQueue class
*/
CircularPriorityQueue::CircularPriorityQueue(size_t lowPrioQueueSize,
                                             size_t normalPrioQueueSize,
                                             size_t highPrioQueueSize,
                                             size_t criticalPrioQueueSize) {
    lowPriorityQueue = new CircularQueue(lowPrioQueueSize);
    normalPriorityQueue = new CircularQueue(normalPrioQueueSize);
    highPriorityQueue = new CircularQueue(highPrioQueueSize);
    criticalPriorityQueue = new CircularQueue(criticalPrioQueueSize);
}

CircularPriorityQueue::~CircularPriorityQueue() {
    delete lowPriorityQueue;
    delete normalPriorityQueue;
    delete highPriorityQueue;
    delete criticalPriorityQueue;
}

/*
** enqueue:
**
** Enqueue inserts an event into each priority queue.
*/
void CircularPriorityQueue::enqueue(const Event& evt) {
    CircularQueue* q = NULL;
    switch (evt.getPriority()) {
        case Event::PRIO_LOW:
            q = lowPriorityQueue;
            break;
        case Event::PRIO_NORMAL:
            q = normalPriorityQueue;
            break;
        case Event::PRIO_HIGH:
            q = highPriorityQueue;
            break;
        case Event::PRIO_CRITICAL:
            q = criticalPriorityQueue;
            break;
    }

    assert(q != NULL);
    q->enqueue(evt);
}

/*
** dequeue:
**
** Dequeues an highest priority event from priority queue.
*/
bool CircularPriorityQueue::dequeue(Event& evt) {
    CircularQueue* q = NULL;

    if (!criticalPriorityQueue->empty()) {
        q = criticalPriorityQueue;
    }
    else if (!highPriorityQueue->empty()) {
        q = highPriorityQueue;
    }
    else if (!normalPriorityQueue->empty()) {
        q = normalPriorityQueue;
    }
    else if (!lowPriorityQueue->empty()) {
        q = lowPriorityQueue;
    }
    else {
        /* Don't use exception here since this condition happens a lot */
        return false;
    }

    assert(q != NULL);
    return q->dequeue(evt);
}

