/*
 * circularqueue.h
 * Copyright 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Steve Grubb <sgrubb@redhat.com>
 *     Junji Kanemaru <junji.kanemaru@linuon.com>
 */
#ifndef CIRCULARQUEUE_H
#define CIRCULARQUEUE_H
#include <string>
#include <pthread.h>
#include "event.h"

/**
 * Class CircularQueue
 * This class holds an array of events. All events are created so that
 * memory fragmentation do not become a problem. Events will use 
 * the assignment operator to acquire values. */
class CircularQueue {
/**
 * Public stuff
 */
public:
    typedef enum {
        ERR_EMPTY = 200
    } CircularQueueError;

	CircularQueue (size_t maxSize = 1024);
    ~CircularQueue();

	/**
	 * This method places event data into the queue. The array of events are 
     * already created. This means that an assignment operator will be used 
     * to add the event to the array. */
	void enqueue (const Event& elem);

	/**
	 * This method removes data from the queue. Because of concerns regarding 
     * dynamic creation and deletion of classes in a long running server, 
     * this will effectivly just zero out the event after copying it to 
     * a stack variable. */
	bool dequeue (Event& elem);

	/**
	 * This returns the total number of ietms in the each queue.
	 */
	size_t count () const { return size; }

    /**
     * This method returns TRUE if queue is empty
     */
    bool empty() const { return (size == 0); }

/**
 * Protected stuff
 */
protected:

/**
 * Private stuff
 */
private:
    size_t maxSize;
    size_t size;
    size_t head, tail;

    pthread_mutex_t mutex;

	/**
	 * This is an array of events. They are already created so that we do not
     * have dynamic class creation or deletion. This is to avoid fragmentation
     * or memory leaks.
	 */
    Event* array;

	/**
	 * Operations
	 */
};


/**
** CircularPriorityQueue class
**
** This class prioritize elements based on Event priority.
** Although there's STL priority_queue there I did very primitive way that
** having CircularQueue for each priority but acts like a single CircularQueue.
** With STL priority_queue, we need to pay cost of Less-Than comparison and
** sorting queue elements.  */
#define DEFALT_LOWPRIO_QUEUE_SIZE 512
#define DEFALT_NORMALPRIO_QUEUE_SIZE 2048
#define DEFALT_HIGHPRIO_QUEUE_SIZE 512
#define DEFALT_CRITICALPRIO_QUEUE_SIZE 256
class CircularPriorityQueue {
public:
    CircularPriorityQueue(
                size_t lowPrioQueueSize = DEFALT_LOWPRIO_QUEUE_SIZE,
                size_t normalPrioQueueSize = DEFALT_NORMALPRIO_QUEUE_SIZE,
                size_t highPrioQueueSize = DEFALT_HIGHPRIO_QUEUE_SIZE,
                size_t criticalPrioQueueSize = DEFALT_CRITICALPRIO_QUEUE_SIZE);

    ~CircularPriorityQueue();

    void enqueue(const Event& evt);
    bool dequeue(Event& evt);

    size_t count() const;
    bool empty() const;
 
private:
    CircularQueue* criticalPriorityQueue;
    CircularQueue* highPriorityQueue;
    CircularQueue* normalPriorityQueue;
    CircularQueue* lowPriorityQueue;
};

inline size_t CircularPriorityQueue::count() const {
    return (criticalPriorityQueue->count() + highPriorityQueue->count() +
            normalPriorityQueue->count() + lowPriorityQueue->count());
}

inline bool CircularPriorityQueue::empty() const {
    return (count() == 0);
}

#endif //CIRCULARQUEUE_H

