/*
 * componentconfigurator.cpp
 * Copyright 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Steve Grubb <sgrubb@redhat.com>
 */
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include "componentconfigurator.h"
#include "application.h"
#include "libaudit.h"
#include "private.h"

/**
 * Constructors/Destructors
 */
ComponentConfigurator *ComponentConfigurator::_instance = 0;

/**
 * Methods
 */
/**
 * Get the value of repo
 * This class owns the Component Repository.
 * @return the value of repo
 */
ComponentRepository ComponentConfigurator::get_repo ( ) {
	return repo;
}
/**
 * Set the value of repo
 * This class owns the Component Repository.
 * @param value the value of repo
 */
void ComponentConfigurator::set_repo (ComponentRepository value ) {
	
}
/**
 * Get the value of config_file
 * This is the location where the config file resides.
 * @return the value of config_file
 */
const char * ComponentConfigurator::get_config_file ( ) {
	return config_file;
}
/**
 * Set the value of config_file
 * This is the location where the config file resides.
 * @param value the value of config_file
 */
void ComponentConfigurator::set_config_file (const char * value ) {
	if (config_file)
		free((void *)config_file);
	config_file = strdup(value);
}
/**
 * This method causes the class to read the config file located by the
 * config_file attribute.
 */
int ComponentConfigurator::read_config () 
{
	int rc, fd;
	struct stat st;
	FILE *f;
	char buf[128];

	if (config_file == NULL)
		config_file = strdup("/etc/audispd.conf");

	rc = open(config_file, O_NOFOLLOW|O_RDONLY);
	if (rc < 0) {
		if (errno != ENOENT) {
			audit_msg(LOG_ERR, "Error opening config file (%s)",
				strerror(errno));
			return 1;
		}
		audit_msg(LOG_WARNING,
			"Config file %s doesn't exist, skipping", config_file);
		return 0;
	}
	fd = rc;

	/* check the file's permissions: owned by root, not world writable,
	 * not symlink.
	 */
	audit_msg(LOG_DEBUG, "Config file %s opened for parsing",
                        config_file);
	if (fstat(fd, &st) < 0) {
		audit_msg(LOG_ERR, "Error fstat'ing config file (%s)",
			strerror(errno));
		return 1;
	}
	if (st.st_uid != 0) {
		audit_msg(LOG_ERR, "Error - %s isn't owned by root",
			config_file);
		return 1;
	}
	if ((st.st_mode & S_IWOTH) == S_IWOTH) {
		audit_msg(LOG_ERR, "Error - %s is world writable",
			config_file);
		return 1;
	}
	if (!S_ISREG(st.st_mode)) {
		audit_msg(LOG_ERR, "Error - %s is not a regular file",
			config_file);
		return 1;
	}

	/* it's ok, read line by line */
	f = fdopen(fd, "r");
	if (f == NULL) {
		audit_msg(LOG_ERR, "Error - fdopen failed (%s)",
			strerror(errno));
		return 1;
	}

	while (get_line(f, buf)) {
		// convert line into name-value pair
	}

	return 0;
}

/**
 * This method will start the init process for the whole program. It
 * should only be invoked once.
 */
int ComponentConfigurator::init ()
{
	int rc;

	// create repo
	rc = read_config();
	if (rc)
		return rc;

	// Create the application component. This will be needed no
	// matter what the config says.
	Application *app = Application::instance();
	app->init("dummy config string");
	repo.insert((Component &)*app);

	// FIXME: create components based on config

    return 0;
}
/**
 * This method tells the configurator to read the config file and
 * reconfigure all components without taking down the application.
 */
void ComponentConfigurator::reconfigure () {
	
}
/**
 * This method is called to stop all processing before terminating
 * the application or reconfiguring the plugins.
 */
void ComponentConfigurator::suspend_all () {
	
}

char *ComponentConfigurator::get_line(FILE *f, char *buf)
{
        if (fgets(buf, 128, f)) {
                /* remove newline */
                char *ptr = strchr(buf, 0x0a);
                if (ptr)
                        *ptr = 0;
                return buf;
        }
        return NULL;
}

