/* auditd-plugin-lib.h --
 * Copyright 2005 Linuon Inc. Tokyo Japan
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Junji Kanemaru <junji.kanemaru@linuon.com>
 * 
 */

#ifndef __AUDITD_PLUGIN_LIB_H__
#define __AUDITD_PLUGIN_LIB_H__

#include <pthread.h>
#include "auditd-plugin.h"

#define MAX_PLUGINS 3

#ifdef __cplusplus
extern "C" {
#endif

/* Structures for plugin combinations which specified in config file
** */
/* Plugin combination structure */
typedef struct _plugin_combo_t {
    char* receiver_plugin;
    char* interpreter_plugin;
    char* consumer_plugin;
} plugin_combo_t;

/* Plugin file list structure */
typedef struct _plugin_file_list_t {
    int cnt;
    plugin_combo_t combo[MAX_PLUGINS];
} plugin_file_list_t;


/* Structures for event processing.
**
** Plugins are specified in auditd.conf. The format is:
**
** Plugin R:receiver_plugin,I:interpreter_plugin,C:consumer_plugin
**
** where *_plugin are plugin file name. The directory path is automatically
** constructed from $Home/plugin. $Home is auditd home directory which is
** also specified in auditd.conf.
** It is possible to use single plugin for all if the plugin supports
** all interface, i.e. let's say you have default_plugin.so then it looks like:
**
** Plugin R:default_plugin.so,I:default_plugin.so,C:default_plugin.so
**
** If you have DBUS output plugin, dbus_plugin.so and want to use it for
** another output of audit event then you would want to have two definition:
**
** Plugin R:default_plugin.so,I:default_plugin.so,C:default_plugin.so
** Plugin R:default_plugin.so,I:default_plugin.so,C:dbus_plugin.so
**
** In this case auditd plugin modules look like this:
**
** receiver_list[0]: default_plugin->receive_event
**  Enqueue raw events to receiver_list[0].raw_event_list.
**
** interpreter_list[0]: default_plugin->interpret_event
**  Dequeue raw events from raw_event_list which is equivalent to
**  receiver's raw_event_list in this configuration, then enqueue 
**  interpreted events to interpreted_event_list.
**
** consumer_list[0]: default_plugin->consume_event
** consumer_list[1]: dbus_plugin->consume_event
**  Both consumer refer same interpreted_event_list according to auditd.conf.
**  
**  
*/
/* Data container. Event data element to store raw and interpreted msg. */
typedef struct _audit_event_elem_t {
    auditd_plugin_type_t    type;
    size_t                  size;
    void*                   data; /* reference, not a copy */
    struct _audit_event_elem_t* next;
} audit_event_elem_t;

/* Event list(also used for interpreted event data) */
typedef struct _audit_event_list_t {
    pthread_mutex_t     lock;
    pthread_cond_t      cond;
    audit_event_elem_t* head;
    audit_event_elem_t* tail;
    unsigned long       size;
} audit_event_list_t;

/* Error handler structure */
typedef struct _error_handler_t {
    on_error_method_t       on_error_callback;
    int                     done;
} error_handler_t;
#define AUDITD_PLUGIN_THREAD_NOT_FINISHED   0
#define AUDITD_PLUGIN_THREAD_FINISHED       1

/* Event receiver list.
** Each receiver has own event list. One method, one event list to write and
** one thread for receiver. */
typedef struct _receiver_list_t {
    pthread_t               pthread_id;
    receive_event_method_t  receive_event;
    error_handler_t         error_handler;
    audit_event_list_t*     raw_event_list;
    plugin_private_storage_t pps; /* for each thread */
    struct _receiver_list_t* next;
} receiver_list_t;

/* Event interpreter list.
** Each interpreter has source event list and output list where
** to store interepreted event. */
typedef struct _interpreter_list_t {
    pthread_t                   pthread_id;
    interpret_event_method_t    interpret_event;
    error_handler_t             error_handler;
    audit_event_list_t*         raw_event_list;
    audit_event_list_t*         interpreted_event_list;
    plugin_private_storage_t    pps; /* for each thread */
    struct _interpreter_list_t* next;
} interpreter_list_t;

/* Event consumer list.
** Each consumer has interepreted event data list. */
typedef struct _cunsumer_list_t {
    pthread_t                   pthread_id;
    consume_event_method_t      consume_event;
    error_handler_t             error_handler;
    audit_event_list_t*         interpreted_event_list;
    plugin_private_storage_t    pps; /* for each thread */
    struct _cunsumer_list_t*    next;
} consumer_list_t;

/* Plugin api list */
typedef struct _plugin_api_list_t {
    receiver_list_t     receiver_list;
    interpreter_list_t  interpreter_list;
    consumer_list_t     consumer_list;
} plugin_api_list_t;


/* External functions
**/
/* init_plugins: */
int init_plugins(
    const auditd_env_t* env, 
    const plugin_file_list_t* plugin_file_list);

/* start_event_threads: */
int start_event_threads(const auditd_env_t* env);

/* Cancel event threads */
void cleanup_plugins(void);

#ifdef __cplusplus
}
#endif

#endif /* __AUDITD_PLUGIN_LIB_H__ */
