/*
 * value.cpp
 * Copyright 2005 Linuon Inc. Tokyo Japan All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Junji Kanemaru <junji.kanemaru@linuon.com>
 */
/* C headers */
#include <stdio.h>
#include <string.h>
#include <assert.h>

/* C++ headers */
#include <vector>
#include <map>
#include <exception>
#include <cassert>

/* Private headers */
#include "value.h"
#include "fault.h"

using namespace std;

/**
** ValueData class (ValueData has already been pre-declared in value.h)
**
** ValueData class is used as placeholder of idata of Value class.
** Value objects could share same ValueData instance, for example
** you have a ValueString object like below:
**
** ValueString valStr("Value String");
**
** and you want to copy ValueString into Value object like:
**
** Value v = valStr;
**
** then ValueData won't copy real value data "Value String", instead make
** a reference to it and increment reference count.
** So right after above line the reference count become 2.
** The data storage will be deleted when the object gets deleted with
** reference count 1.
*/
class Value::ValueData {
public:
    ValueData();
    ValueData(size_t size, const void* data = NULL);
    ~ValueData();

    const void* contents()  const { return _data; }
    size_t size() const { return _size; }

    void incRef();
    void decRef();
    int refCnt()  const { return _refCnt; }

private:
    int _refCnt;    /* Reference count of _data */
    size_t _size;   /* Size of buffer of _data */

    void* _data;    /* This pointer would be referenced with other ValueData
                    ** objects. We won't free _data until there's no object
                    ** referencing it. */
};

Value::ValueData::ValueData() {
    _size = 0;
    _data = NULL;
    _refCnt = 1;
}

Value::ValueData::ValueData(size_t size, const void* data) {
    _data = malloc(size);
     assert(_data != NULL);

    memset(_data, 0, size);

    /* If source is set */
    if (data) {
        memcpy(_data, data, size);
    }

    _size = size;
    _refCnt = 1;
}

Value::ValueData::~ValueData() {
    decRef();
}

inline void Value::ValueData::incRef() {
    _refCnt++;
}

inline void Value::ValueData::decRef() {
     assert(_refCnt > 0);

    _refCnt--;
    if (_refCnt == 0) {
        if (_data) {
            free(_data);
        }
        _size = 0;
        _data = NULL;
    }
    else if (_refCnt < 0) {
        _refCnt = 0;
    }
}

/**
** Base Value class
**
** This class is a abstracted container that hides type kind.
** For instance, if you want to have a stack class which can be pushed
** any kind of data like int, void*, std::string, etc. then
** template won't help, it allows only same type.
** We needed to have totally abstracted class to manage abstracted type.
** This class has field "ValueData* _data", yes it is pointer, and
** Value objects share it. Value objects could refer same ValueData object.
** It prevents redundant object creation and copying. It is OK because
** Vlaue obejct is actually container, mostly used as read-only object.
** It is ugly though. I take the blame... -- Junji
*/
const Value Value::VOID_VALUE;

Value::Value() {
    _type = Value::VOID;
    _data = new ValueData;
}

Value::Value(ValueType type, size_t size, const void* data) {
    _type = type;
    _data = new ValueData(size, data);
}

Value::Value(const Value& v) {
    _type = v._type;
    _data = v._data;
    incRef();
}

Value::~Value() {
    decRef();
}

Value& Value::operator =(const Value& v) {
    assert(_type != DEAD);
    assert(v._type != DEAD);

    /* Do decRef current this->_data first. */
    decRef();

    /* Then copy Vlaue into this. We incRef after copying fields because
    ** v.incRef() discards qualifiers of const Value&,
    ** a bit hackish though...  */
    _type = v._type;
    _data = v._data;
    incRef();

    return *this;
}

bool Value::operator ==(const Value& v) const {
    return (_type == v._type && _data == v._data);
}

void Value::decRef() {
    assert(_type != DEAD);

    /* If _data->refCnt == 1 then there's no other object is referencing
    ** _data so we can release it. */
    int refCnt = _data->refCnt();

    /* refCnt must be equal or greater than 1. Otherwise there's 
    ** something wrong */
    assert(refCnt >= 1);

    if (refCnt == 1) {
        switch (_type) {
            /* These cases we just delete "_data" later(below) */
            case VOID:
            case BOOL:
            case INT32:
            case STRING:
            case PTR:
                break;
            case ARRAY: {
                vector<Value>** vpp = (vector<Value>**)_data->contents();
                delete *vpp;
                break;
            }
            case STRUCT: {
                map<string, Value>** spp = 
                        (map<string, Value>**)_data->contents();
                delete *spp;
                break;
            }
            case DEAD:
            default:
                throw APP_FAULT(DEAD_VALUE, "DEAD_VALUE: Value::decRef()");
        }

        /* Release ValueData* _data object */
        delete _data;
        _data = NULL;
        _type = DEAD;
    }
    else {
        _data->decRef();
    }
}

void Value::incRef() {
    assert(_type != DEAD);
    _data->incRef();
}

int Value::refCnt() const {
    assert(_type != DEAD);
    return _data->refCnt();
}

Value& Value::operator [](const int idx) const {
    assert(_type != DEAD);

    if (_type != ARRAY) {
        throw APP_FAULT(FATAL, 
            string("FATAL: Invalid operator [](int index) with type ") +
            getTypeString());
    }

    vector<Value>** vpp = (vector<Value>**)_data->contents();
    if ((size_t)(idx + 1) > (**vpp).size()) {
        throw APP_FAULT(INDEX_OUT_OF_BOUND, "INDEX_OUT_OF_BOUND");
    }
    return (**vpp)[idx];
}

Value& Value::operator [](const string& key) const {
    assert(_type != DEAD);

    if (_type != STRUCT) {
        throw APP_FAULT(FATAL,
            string("FATAL: Invalid operator [](string key) with type ") +
            getTypeString());
    }

    map<string, Value>** spp = (map<string, Value>**)_data->contents();
    map<string, Value>::iterator it = (**spp).find(key);
    if (it == (**spp).end()) {
        throw APP_FAULT(NO_SUCH_KEY_FOUND, "NO_SUCH_KEY_FOUND");
    }

    /* C++ STL map.find() returns <pair> */
    return (*it).second;
}

Value& Value::operator [](const char* key) const {
    return operator [](string(key));
}

bool Value::getBool() const {
    assert(_type != DEAD);

    if (_type != BOOL) {
        throw APP_FAULT(INVALID_TYPE_CAST, "INVALID_TYPE_CAST");
    }
    return *((const bool*)_data->contents());
}

int Value::getInt() const {
    assert(_type != DEAD);

    if (_type != INT32) {
        throw APP_FAULT(INVALID_TYPE_CAST, "INVALID_TYPE_CAST");
    }
    return *((const int*)_data->contents());
}

const char* Value::getCString() const {
    assert(_type != DEAD);

    if (_type != STRING) {
        throw APP_FAULT(INVALID_TYPE_CAST, "INVALID_TYPE_CAST");
    }
    return (const char*)_data->contents();
}

string Value::getString() const {
    assert(_type != DEAD);

    if (_type != STRING) {
        throw APP_FAULT(INVALID_TYPE_CAST, "INVALID_TYPE_CAST");
    }
    return string((const char*)_data->contents());
}

const void* Value::getPtr() const {
    assert(_type != DEAD);

    if (_type != PTR) {
        throw APP_FAULT(INVALID_TYPE_CAST, "INVALID_TYPE_CAST");
    }
    return (const void*)_data->contents();
}

Value::ValueType Value::type() const {
    assert(_type != DEAD);

    return _type;
}

size_t Value::size() const {
    assert(_type != DEAD);

    switch (_type) {
        case VOID:
            return 0;
        case BOOL:
            return sizeof(bool);
        case INT32:
            return sizeof(long int);
        case PTR:
            return _data->size();
        case STRING:
            return _data->size() - 1;
        case ARRAY: {
            vector<Value>** vpp = (vector<Value>**)_data->contents();
            return (**vpp).size();
        }
        case STRUCT: {
            map<string, Value>** spp = (map<string, Value>**)_data->contents();
            return (**spp).size();
        }
        case DEAD:
            throw APP_FAULT(DEAD_VALUE, "DEAD_VALUE: Value::size()");
        default:
            throw APP_FAULT(FATAL, "unknow type");
    }
}

string Value::getTypeString() const {
    switch(_type) {
        case VOID:
            return string("VOID");
        case BOOL:
            return string("BOOL");
        case INT32:
            return string("INT32");
        case STRING:
            return string("STRING");
        case ARRAY:
            return string("ARRAY");
        case STRUCT:
            return string("STRUCT:");
        case PTR:
            return string("PTR");
        case DEAD:
            return string("DEAD");
        default:
            break;
    }

    return string("Unknown type");
}

/**
** ValueBoolean class */
ValueBoolean::ValueBoolean(bool b) : Value(BOOL, sizeof(bool), &b) {
}

/**
** ValueInt32 class */
ValueInt32::ValueInt32(int i) : Value(INT32, sizeof(long int), &i) {
}

/**
** ValueString class */
ValueString::ValueString(const string& str)
        : Value(STRING, str.length() + 1, (const void*)str.c_str()) {
}

ValueString::ValueString(const char* str) 
        : Value(STRING, strlen(str) + 1, (const void*)str) {
}

/**
** ValueArray class */
ValueArray::ValueArray() : Value(ARRAY, sizeof(vector<Value>**)) {
    vector<Value>** vpp = (vector<Value>**)valueData().contents();
    *vpp = new vector<Value>;
}

void ValueArray::append(const Value& v) {
    vector<Value>** vpp = (vector<Value>**)valueData().contents();
    (**vpp).push_back(v);
}

void ValueArray::remove(const Value& v) {
    vector<Value>** vpp = (vector<Value>**)valueData().contents();
    vector<Value>::iterator it;
    for (it = (**vpp).begin(); it != (**vpp).end(); it++) {
        if ((*it) == v) {
            (**vpp).erase(it);
            break;
        }
    }
}

void ValueArray::operator +=(const Value& v) {
    append(v);
}

void ValueArray::operator -=(const Value& v) {
    remove(v);
}

/**
** ValueStruct class */
ValueStruct::ValueStruct() : Value(STRUCT, sizeof(map<string, Value>**)) {
    map<string, Value>** spp = (map<string, Value>**)valueData().contents();
    *spp = new map<string, Value>;
}

void ValueStruct::set(const std::string& key, const Value& v) {
    map<string, Value>** spp = (map<string, Value>**)valueData().contents();
    (**spp)[key] = v;
}

void ValueStruct::set(const char* key, const Value& v) {
    set(string(key), v);
}

Value& ValueStruct::get(const std::string& key) {
    map<string, Value>** spp = (map<string, Value>**)valueData().contents();
    return (**spp)[key];
}

Value& ValueStruct::get(const char* key) {
    return get(string(key));
}

void ValueStruct::erase(const std::string& key) {
    map<string, Value>** spp = (map<string, Value>**)valueData().contents();
    map<string, Value>::iterator it = (**spp).find(key);
    if (it == (**spp).end()) {
        throw APP_FAULT(NO_SUCH_KEY_FOUND, "NO_SUCH_KEY_FOUND");
    }

    (**spp).erase(it);
}

void ValueStruct::erase(const char* key) {
    erase(string(key));
}

/**
** ValuePtr class */
ValuePtr::ValuePtr(size_t size, const void* ptr) : Value(PTR, size, ptr) {
}

